#include <iostream>
#include <iomanip>
#include <queue>
#include <stdio.h>
#include <cstring>
#include <stdlib.h>
#include "disass.h"

using namespace std;

unsigned int loadAddress;
unsigned int endAddress;
size_t codeSize;
daDescriptor daMap[65536 + 2];
queue<unsigned int> pcStack;

static bool cbmisprint(unsigned char &cbmchar)
{
	//if (cbmchar >= 0x01 && cbmchar <= 0x1A) {
	//	cbmchar ^= 0x40;
	//	return true;
	//}
	//if (cbmchar >= 0x41 && cbmchar <= 0x5A) {
	//	cbmchar ^= 0x20;
	//	return true;
	//}
	//if (cbmchar >= 0x30 && cbmchar <= 0x39)
	//	return true;
	if (isprint((char) cbmchar))
		return true;
	return false;
}

void initMap(unsigned short from, unsigned short to, const unsigned char *mem, DaOptions& options)
{
    unsigned short startAddress;

    for(unsigned int i = 0; i < 65536; i++) {
        daMap[i].address = -1;
        daMap[i].addrMode = NUMOPS - 1;
        daMap[i].opLength = 1;
        daMap[i].content = CT_DATA;
        daMap[i].outOfBounds = !(i >= from && i <= to);
        daMap[i].opcode = (i >= from && i <= to) ? mem[i - from] : 0;
        daMap[i].emitLabel = 0;
    }
    pcStack.empty();
    // CBM ROM's have 2 entry addresses
    if (!strncmp((const char *) mem + 7, "CBM", 3)) {
        pcStack.push(loadAddress + 3);
        daMap[loadAddress + 3].emitLabel = 1;
    }
    // default start address is load address
    startAddress = loadAddress;
    char *sysptr = 0;

	while (options.startAddressList.size()) {
		unsigned short sa = options.startAddressList.back();
		if (!daMap[sa].outOfBounds) {
			pcStack.push(sa);
			options.startAddressList.pop_back();
			daMap[sa].emitLabel = 1;
		}
	}

    if (loadAddress == 0x1000 || loadAddress == 0x0800 || loadAddress == 0x1001 || loadAddress == 0x0801) {
        unsigned int i = 32;
        do {
            // look for SYS token
            if (mem[i] == 0x9e) {
                sysptr = (char*)mem + i + 1;
                break;
            }
        } while (--i);
    }
    if (sysptr) {
        unsigned short sa = atoi(sysptr);
		// ccould be some leftover BASIC code pointing to out of bounds address
		if (!daMap[sa].outOfBounds) {
			startAddress = sa;
		}
		pcStack.push(startAddress);
		daMap[startAddress].emitLabel = 1;
	}
    // presumably a system/kernal ROM with reset and IRQ vectors
    if (loadAddress + codeSize == 0x10000) {
        unsigned short address;
        // RESET vector
        address = mem[0xfffe - loadAddress] | (mem[0xffff - loadAddress] << 8);
        pcStack.push(address);
        daMap[address].emitLabel = 1;
        // IRQ vector
        address = mem[0xfffc - loadAddress] | (mem[0xfffd - loadAddress] << 8);
        pcStack.push(address);
        daMap[address].emitLabel = 1;
        // NMI vector
        address = mem[0xfffa - loadAddress] | (mem[0xfffb - loadAddress] << 8);
		if (!daMap[address].outOfBounds) {
			pcStack.push(address);
		}
		daMap[address].emitLabel = 1;
    }
	// no BASIC start SYS code found
    if (loadAddress == startAddress)
    {
        pcStack.push(startAddress);
        daMap[startAddress].emitLabel = 1;
    }
}

void emitDisassembly(DaOptions& options)
{
    char formatTxt[256];
    char daTxt[256] = "";
    char addrTxt[256] = "";
    unsigned int dataCount = 0;

    if (options.fmt) {
        // emit header
        cout << "   ORG $"<< setfill('0') << hex << setw(loadAddress < 0x100 ? 2 : 4) << loadAddress << " - 2" << endl;
		cout << "   DW $"<< setfill('0') << hex << setw(loadAddress < 0x100 ? 2 : 4) << loadAddress << endl;

        // first dump out of bounds references
        for(unsigned int PC = 0; PC < 65536; PC++) {
            if (daMap[PC].outOfBounds && daMap[PC].emitLabel) {
                unsigned short a = PC;
                cout << "L_" << setfill('0') << hex << setw(a < 0x100 ? 2 : 4) << uppercase
                    << a << " = $" << setw(4) << a << " ; (" << dec << daMap[PC].emitLabel << ")" << endl;
            }
        }
        // now dump code
        for(unsigned int PC = 0; PC < 65536; ) {
            if (!daMap[PC].outOfBounds) {

                // finish pending data dump
                if (dataCount && (daMap[PC].content == CT_CODE || daMap[PC].emitLabel)) {
                    cout << endl;
                    dataCount = 0;
                }
                // code labels first
                if (daMap[PC].emitLabel)
                    cout << "L_"<< setfill('0') << hex << setw(4) << uppercase
                        << PC << " ";

                if (!daMap[PC].emitLabel && !dataCount) {
                    cout << "       ";
                }

                if (daMap[PC].content == CT_CODE) {

                    // write mnemonic
                    sprintf(daTxt, "%s ", optable[daMap[PC].opcodeIndex].name);
                    cout << daTxt;

                    unsigned short address = daMap[PC].address;
                    switch (daMap[PC].addrMode) {
                        case AM_IMD:
                            sprintf(formatTxt, "#$%02X", address);
                            break;
                        case AM_IMP:
                            sprintf(formatTxt, "");
                            break;
                        case AM_ZPX:
                        case AM_ZPY:
                        case AM_ZP:
                        case AM_INX:
                        case AM_INY:
                            sprintf(formatTxt, "L_%02X", address);
                            break;
                        case AM_IND:
                            sprintf(formatTxt, "(L_%04X)", address);
                            break;
                        default:
                            sprintf(formatTxt, "L_%04X", address);
                            break;
                    }
                    cout << formatTxt << endl;

                } else {
                    if (dataCount == 16 || !(PC & 0x0F)) {
                        dataCount = 0;
                        cout << endl;
						if (!(PC & 0x007F)) {
							cout << "N_"<< setfill('0') << hex << setw(4) << uppercase << PC << " ";
						} else {
							cout << "       ";
						}
                    }
                    if (!dataCount++) {
                        cout << "DB ";
                    } else {
                        cout << ", ";
                    }
                    unsigned char opcode = daMap[PC].opcode;
                    //cout << "$" << setw(2) << setfill('0') << hex << opcode;
                    if (opcode == '\"' || !isalpha(opcode)) {
                        sprintf(formatTxt, "$%02X", opcode);
                    } else {
                        sprintf(formatTxt, "\"%c\"", opcode);
                    }
                    cout << formatTxt;
                }
                PC += daMap[PC].opLength;
            } else {
                PC += 1;
            }
        }
        cout << endl << "; EOF " << endl;
    } else {
        // disassembly in TEDMON format
        for(unsigned int PC = 0; PC < 65536;) {

            unsigned short address = daMap[PC].address;
            switch (daMap[PC].addrMode) {
                case AM_IMP:
                    break;
                case AM_REL:
                case AM_IMD:
                case AM_ZPX:
                case AM_ZPY:
                case AM_ZP:
                case AM_ABX:
                case AM_ABY:
                case AM_ABS:
                case AM_IND:
                case AM_INX:
                case AM_INY:
                    sprintf(addrTxt, fstrings[daMap[PC].addrMode], address);
                    break;
            }

            switch (daMap[PC].opLength) {
                case 1:
                default:
                    sprintf(formatTxt, "     ");
                    break;
                case 3:
                    sprintf(formatTxt, "%02X %02X", address & 0xff , address >> 8);
                    break;
                case 2:
                    sprintf(formatTxt, "%02X   ", address);
            }

            unsigned int ocix = daMap[PC].opcodeIndex;
            sprintf(daTxt, ". %04X %02X %s %s %s", PC,
                    daMap[PC].opcode, formatTxt, optable[ocix].name, addrTxt);

            if (!daMap[PC].outOfBounds)
                cout << daTxt << endl;
            PC += daMap[PC].opLength;
        }
    }
}

static int findIndex(unsigned char opcode, unsigned int &opix)
{
    opix = NUMOPS - 1;
    for(unsigned int i = 0; i < NUMOPS; i++) {
        unsigned int j = 0;
        int k;
        do {
            k = optable[i].modes[j];
            if ((k & 0xff) == opcode) {
                opix = i;
                return k;
            }
            j++;
        } while(k != -1);
    }
    return -1;
}

int analyze(unsigned char *memory, size_t size, unsigned int flowMode = 1)
{
    unsigned short i;
    unsigned char opcode;
    unsigned short address = 0;
    unsigned short PC;

    PC = (unsigned short) pcStack.front();
    pcStack.pop();
	cerr << "Popping stack PC: " << hex << PC << endl;
    i = PC - loadAddress;

    do {
        unsigned int opCodeIx;
        opcode = memory[i];
        int ix = findIndex(opcode, opCodeIx);
        unsigned int addrMode = (ix >> 16);
        unsigned int opLen = flengths[addrMode] + 1;

        daMap[PC].addrMode = addrMode;
        daMap[PC].opLength = opLen;
        daMap[PC].opcodeIndex = opCodeIx;

        // mark as code if not illegal opcode
        if (opCodeIx < NUMOPS-1 ) {
            for (unsigned int l = 0; l < opLen && (i + l <= size); l++)
                daMap[PC + l].content = CT_CODE;
        }

        // get addressing mode
        switch (addrMode) {
            case AM_IMP:
            case AM_IMD:
                address = memory[(i + 1) & 0xffff];
                break;
            case AM_REL:
                address = memory[(i + 1) & 0xffff];
                address = PC + 2 + (signed char) address;
                daMap[address].emitLabel += 1;
                break;
            case AM_ZPX:
            case AM_ZPY:
            case AM_ZP:
            case AM_INX:
            case AM_INY:
                address = memory[(i + 1) & 0xffff];
                daMap[address].emitLabel += 1;
                break;
            case AM_ABX:
            case AM_ABY:
            case AM_ABS:
				address = memory[(i + 1) & 0xffff] | (memory[(i + 2) & 0xffff] << 8);
                daMap[address].emitLabel += 1;
				break;
            case AM_IND:
                address = memory[(i + 1) & 0xffff] | (memory[(i + 2) & 0xffff] << 8);
				daMap[address].emitLabel += 1;
                break;
            case -1:
            default:
				;
                //PC++;
                //continue;
        }
        daMap[PC].address = address;
        //controlFlow(
        if (flowMode) {
            switch (optable[opCodeIx].isbranch) {
                default:
                    i += opLen;
                    break;
                case 2:
                    // push return address to stack
                    pcStack.push(PC + opLen);
                case 1:
                    // jump to new address
					if (!daMap[address].outOfBounds && addrMode == AM_IND) {
						address = memory[(address - loadAddress + 1) & 0xffff]
						| (memory[(address - loadAddress + 2) & 0xffff] << 8);
					}
					if (!daMap[address].outOfBounds) {
						i = address - loadAddress;
					}
                    break;
                case 3:
				case 4:
                    // get return address from stack
					if (pcStack.empty()) {
						cerr << "Address queue depleted at: $" << PC << endl;
                        return 0;
					}
                    i = pcStack.front() - loadAddress;
                    pcStack.pop();
					cerr << "Popping stack PC: " << hex << loadAddress + i << endl;
                    break;
            }
        } else {
            // dumb disassembly of everything
            i += opLen;
        }
        PC = loadAddress + i;
        // see if we reached a code block or reached bounds
        if (daMap[PC].outOfBounds || daMap[PC].content == CT_CODE) {
            if (!pcStack.empty()) {
                i = pcStack.front() - loadAddress;
				pcStack.pop();
                PC = loadAddress + i;
                cerr << "Popping stack PC: " << hex << PC << endl;
            } else {
				cerr << "Address queue depleted at: $" << PC << endl;
                return 0;
            }
        }
	} while(1); // && i <= size !pcStack.empty()

    return 0;
}

int disass(unsigned char *memory, size_t size, DaOptions& options)
{
    loadAddress = memory[0] | ( memory [1] << 8);
    codeSize = size - 2;
    endAddress = loadAddress + codeSize - 1;

    cerr << "Size        : " << size - 2 << " bytes" << endl;
    cerr << "Load address: $" << hex << uppercase << setw(4) << loadAddress << endl;
    cerr << "End address : $" << hex << uppercase << setw(4) << endAddress << endl;

    initMap(loadAddress, endAddress, memory + 2, options);
    analyze(memory + 2, codeSize, 0);
    emitDisassembly(options);
    return 0;
}
