#ifndef _DISASS_H
#define _DISASS_H

#include <vector>

enum DaOutputFmt {
	DAOF_TEDMON,
	DAOF_AS65,
	DAOF_TASM,
	DAOF_DASM
};

struct DaOptions {
	DaOutputFmt fmt;
	unsigned int verbosity;
	bool lowercase;
	std::vector<unsigned short> startAddressList;
};

enum ContentType {
    CT_DATA,
    CT_CODE
};

struct daDescriptor {
    ContentType content;
    unsigned int emitLabel;
    bool outOfBounds;
    unsigned char opcode;
    unsigned int opcodeIndex;
    unsigned short address;
    int addrMode;
    int opLength;
};

extern int disass(unsigned char *memory, size_t size, DaOptions& options);

static const char *fstrings[12]=
{
        "#$%02X",       // immediate
        "$%04X",        // RELATIVE(jump)
        "$%02X",        // Z
        "$%02X,X",      // Z,x
        "$%02X,Y",      // Z,y
        "$%04X",        //ABS
        "$%04X,X",      // ABS,x
        "$%04X,Y",      // ABS,y
        "($%04X)",      // IND
        "($%02X,X)",    // INX
        "($%02X),Y",    // INY
	""
};
static const int flengths[12]={1,1,1,1,1,2,2,2,2,1,1,0};


#define AM_IMD  (0)
#define AM_REL  (1)
#define AM_ZP   (2)
#define AM_ZPX  (3)
#define AM_ZPY  (4)
#define AM_ABS  (5)
#define AM_ABX  (6)
#define AM_ABY  (7)
#define AM_IND  (8)
#define AM_INX  (9)
#define AM_INY  (10)
#define AM_IMP  (11)

#define IMD(x)	((AM_IMD<<16)|x)
#define	REL(x)	((AM_REL<<16)|x)
#define	ZP(x)	((AM_ZP<<16)|x)
#define	ZPX(x)	((AM_ZPX<<16)|x)
#define	ZPY(x)	((AM_ZPY<<16)|x)
#define	ABS(x)	((AM_ABS<<16)|x)
#define	ABX(x)	((AM_ABX<<16)|x)
#define ABY(x)	((AM_ABY<<16)|x)
#define	IND(x)	((AM_IND<<16)|x)
#define	INX(x)	((AM_INX<<16)|x)
#define	INY(x)	((AM_INY<<16)|x)
#define IMP(x)	((AM_IMP<<16)|x)

typedef struct {
	const char name[4];
	int isillegal;
    int isbranch; // 0: not 1: simple jump 2: branching 3: -- 4: return from sub
	int type; // 1: read, 2: write, 3: RWMW
	int modes[10];
} OPS;

#define NUMOPS (sizeof(optable)/sizeof(optable[0]))
static OPS optable[]=
{
    {"BRK", 0,4, 2,{IMP(0x00),-1}},
    {"RTI", 0,4, 2,{IMP(0x40),-1}},
    {"RTS", 0,3, 0,{IMP(0x60),-1}},
    {"PHA", 0,0, 2,{IMP(0x48),-1}},
    {"PHP", 0,0, 2,{IMP(0x08),-1}},
    {"PLA", 0,0, 1,{IMP(0x68),-1}},
    {"PLP", 0,0, 1,{IMP(0x28),-1}},
    {"JMP", 0,1, 0,{ABS(0x4C),IND(0x6C),-1}},
    {"JSR", 0,2, 0,{ABS(0x20),-1}},
    {"TAX", 0,0, 0,{IMP(0xAA),-1}},
    {"TXA", 0,0, 0,{IMP(0x8A),-1}},
    {"TAY", 0,0, 0,{IMP(0xA8),-1}},
    {"TYA", 0,0, 0,{IMP(0x98),-1}},
    {"TSX", 0,0, 0,{IMP(0xBA),-1}},
    {"TXS", 0,0, 0,{IMP(0x9A),-1}},
    {"DEX", 0,0, 0,{IMP(0xCA),-1}},
    {"DEY", 0,0, 0,{IMP(0x88),-1}},
    {"INX", 0,0, 0,{IMP(0xE8),-1}},
    {"INY", 0,0, 0,{IMP(0xC8),-1}},
    {"CLC", 0,0, 0,{IMP(0x18),-1}},
    {"CLD", 0,0, 0,{IMP(0xD8),-1}},
    {"CLI", 0,0, 0,{IMP(0x58),-1}},
    {"CLV", 0,0, 0,{IMP(0xB8),-1}},
    {"SEC", 0,0, 0,{IMP(0x38),-1}},
    {"SED", 0,0, 0,{IMP(0xF8),-1}},
    {"SEI", 0,0, 0,{IMP(0x78),-1}},
    {"NOP", 0,0, 0,{IMP(0xEA),-1}},
    {"ASL", 0,0, 1,{IMP(0x0a),ZP(0x06),ZPX(0x16),ABS(0x0E),ABX(0x1E),-1}},
    {"DEC", 0,0, 3,{ZP(0xc6),ZPX(0xd6),ABS(0xcE),ABX(0xdE),-1}},
    {"INC", 0,0, 3,{ZP(0xe6),ZPX(0xf6),ABS(0xeE),ABX(0xfE),-1}},
    {"LSR", 0,0, 3,{IMP(0x4a),ZP(0x46),ZPX(0x56),ABS(0x4E),ABX(0x5E),-1}},
    {"ROL", 0,0, 3,{IMP(0x2a),ZP(0x26),ZPX(0x36),ABS(0x2E),ABX(0x3E),-1}},
    {"ROR", 0,0, 3,{IMP(0x6a),ZP(0x66),ZPX(0x76),ABS(0x6E),ABX(0x7E),-1}},
    {
        "ADC", 0,0, 1,{
            IMD(0x69),ZP(0x65),ZPX(0x75),ABS(0x6D),ABX(0x7d),ABY(0x79),
            INX(0x61),INY(0x71),-1
        }
    },
    {
        "AND", 0,0, 1,{
            IMD(0x29),ZP(0x25),ZPX(0x35),ABS(0x2D),ABX(0x3d),ABY(0x39),
            INX(0x21),INY(0x31),-1
        }
    },
    {"BIT", 0,0, 1,{ZP(0x24),ABS(0x2c),-1}},
    {
        "CMP", 0,0, 1,{
            IMD(0xc9),ZP(0xc5),ZPX(0xd5),ABS(0xcD),ABX(0xdd),ABY(0xd9),
            INX(0xc1),INY(0xd1),-1
        }
    },
    {"CPX", 0,0, 1,{IMD(0xe0),ZP(0xe4),ABS(0xec),-1}},
    {"CPY", 0,0, 1,{IMD(0xc0),ZP(0xc4),ABS(0xcc),-1}},
    {
        "EOR", 0,0, 1,{
            IMD(0x49),ZP(0x45),ZPX(0x55),ABS(0x4D),ABX(0x5d),ABY(0x59),
            INX(0x41),INY(0x51),-1
        }
    },
    {
        "LDA", 0,0, 1,{
            IMD(0xa9),ZP(0xa5),ZPX(0xb5),ABS(0xaD),ABX(0xbd),ABY(0xb9),
            INX(0xa1),INY(0xb1),-1
        }
    },
    {"LDX", 0,0,1,{IMD(0xa2),ZP(0xa6),ZPY(0xB6),ABS(0xae),ABY(0xbe),-1}},
    {"LDY", 0,0,1,{IMD(0xa0),ZP(0xa4),ZPX(0xB4),ABS(0xac),ABX(0xbc),-1}},
    {
        "ORA", 0,0, 1,{
            IMD(0x09),ZP(0x05),ZPX(0x15),ABS(0x0D),ABX(0x1d),ABY(0x19),
            INX(0x01),INY(0x11),-1
        }
    },
    {
        "SBC", 0,0, 1,{
            IMD(0xEB),IMD(0xe9),ZP(0xe5),ZPX(0xf5),ABS(0xeD),ABX(0xfd),ABY(0xf9),
            INX(0xe1),INY(0xf1),-1
        }
    },
    {
        "STA", 0,0, 2,{
            ZP(0x85),ZPX(0x95),ABS(0x8D),ABX(0x9d),ABY(0x99),
            INX(0x81),INY(0x91),-1
        }
    },
    {"STX", 0,0, 2,{ZP(0x86),ZPY(0x96),ABS(0x8E),-1}},
    {"STY", 0,0, 2,{ZP(0x84),ZPX(0x94),ABS(0x8C),-1}},
    {"BCC", 0,2, 1,{REL(0x90),-1}},
    {"BCS", 0,2, 1,{REL(0xb0),-1}},
    {"BEQ", 0,2, 1,{REL(0xf0),-1}},
    {"BNE", 0,2, 1,{REL(0xd0),-1}},
    {"BMI", 0,2, 1,{REL(0x30),-1}},
    {"BPL", 0,2, 1,{REL(0x10),-1}},
    {"BVC", 0,2, 1,{REL(0x50),-1}},
    {"BVS", 0,2, 1,{REL(0x70),-1}},
    {"???", 1,0, 1,{IMP(0x02),-1}}
};

#endif // _DISASS_H
