#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <string>
#include <stdlib.h>
#include "disass.h"

using namespace std;
DaOptions options;

int readFile(const string name, unsigned char* &buffer, size_t& size)
{
    ifstream f(name.c_str(), ios::in|ios::binary|ios::ate);

    if (f.is_open()) {
        size = f.tellg();
		buffer = new unsigned char[size];
        f.seekg (0, ios::beg);
        f.read ((char*) buffer, streamsize(size));
        cerr << "File '" << name.c_str() << "' read successfully." << endl;
        f.close();
        return 1;
    }
    return 0;
}

void printUsage()
{
	cout << "YASDA 1.0 - Yet Another 6502 Disassembler" << endl;
	cout << "Usage: yasda <PRG file> [options]" << endl;
	cout << "Options:" << endl;
	cout << "-a:address1,address2,$hexaddress,$hexaddress - define entry points" << endl;
	cout << "-f:x - set output format to 'x'" << endl;
	cout << "-h  - print this help" << endl;
	cout << "-v  - verbose mode" << endl;
}

int main(unsigned int argc, char **argv)
{
	unsigned char *buffer;
    int retval;
	size_t size;
	//outfile

    if (argc <= 1) {
        cerr << "Missing argument." << endl;
		printUsage();
        return 1;
    }
	std::string inputFileName(argv[1]);
//#ifdef _DEBUG
	//return readFile("C16BASIC.PRG");
	//return readFile("Aardvark.prg");
	//return readFile("C16rom.prg");
	//return readFile("C16kern.prg");
	//return readFile("testCharSet.prg");
	//return readFile("winter_g7ec7.prg", buffer, size);
//#endif
	//inputFileName = "sizapr-u103.bin.prg";
	inputFileName = argv[1];

	// default options
	options.fmt = DAOF_AS65;
	options.verbosity = 0;
	options.startAddressList.empty();

	for(unsigned int j = 2; j < argc; j++) {
		string carg(argv[j]);
		if (carg.substr(0,3) == "-a:") {
			if (argv[j] + 3) {
				stringstream ss(argv[j] + 3);
				while(ss.good()) {
					string arguments;
					stringstream temp;
					unsigned short addr;

					getline(ss, arguments, ',');
					if (arguments.find("$", 1)) {
						temp << hex << arguments.substr(1);
					} else {
						temp << arguments;
					}
					temp >> addr;
					options.startAddressList.push_back(addr);
				}
			}
		} else if (carg.substr(0,2) == "-h") {
			printUsage();
			return 0;
		} else if (carg.substr(0,2) == "-f:" && (argv[j] + 3)) {
			options.fmt = DaOutputFmt(atoi(argv[j] + 3));
		} else if (carg.substr(0,2) == "-v") {
			options.verbosity = 1;
		}
	}
	if ((retval = readFile(argv[1], buffer, size)) ) {
		disass(buffer, size, options);
		delete [] buffer;
		return retval;
	}
    return 0;
}
